package MARC::diffSerials;

use 5.008000;
use strict;
use warnings;
use Carp;

require Exporter;

our @ISA = qw(Exporter);

=head1 NAME

MARC::diffSerials - 図書館目録における雑誌包括所蔵形のデータをもとに、
同一雑誌タイトルについての２館間の所蔵を比較する

=head1 SYNOPSIS

=head2 関数スタイルのインターフェイス

  use MARC::diffSerials;
  # 入力データ（雑誌所蔵編集形）
  $vols_of_LibA = "3(),4-6,9,10()";  # 図書館Ａの雑誌所蔵
  $vols_of_LibB = "2-9";             # 図書館Ｂの雑誌所蔵
  
  # 図書館Ｂが図書館Ａに補充可能な巻を表示する
  print MARC::diffSerials::fill_all($vols_of_LibA, $vols_of_LibB), "\n";
   
  # 図書館Ｂが図書館Ａに補充可能な巻のうち,
  # 図書館Ａの最古巻より古い巻を表示する
  print MARC::diffSerials::fill_old($vols_of_LibA, $vols_of_LibB), "\n";
    
  # 図書館Ｂが図書館Ａに補充可能な巻のうち,
  # 図書館Ａの最新巻より新しい巻を表示する
  print MARC::diffSerials::fill_new($vols_of_LibA, $vols_of_LibB), "\n";
  
  # 図書館Ｂが図書館Ａに補充可能な途中欠号の巻を表示する
  print MARC::diffSerials::fill_middle($vols_of_LibA, $vols_of_LibB), "\n";
   
  # 図書館Ａの途中欠号を表示
  print MARC::diffSerials::middle_lack($vols_of_LibA), "\n";

  # 図書館Ａと図書館Ｂの雑誌所蔵をマージ
  print MARC::diffSerials::marge_vols($vols_of_LibA, $vols_of_LibB), "\n";

=head2 オブジェクト指向のインターフェイス (未実装)

I< 現在は未実装。将来のバージョンにて実装する可能性あり。 >

=head1 DESCRIPTION

図書館目録における雑誌包括所蔵形のデータをもとに、同一雑誌タイトルについての２館間の所蔵を比較する。
なお、使用にあたり次については留意すること。

=over

=item *
２館とも同じ巻(Vol)が不完全巻の場合、 I< Vol() > 例 5()の形式にて、号(No)を明記しない

=item *
もし、最古号もしくは最新号が不完全巻であった場合、それは中途欠欠号に含める。逆に所蔵より古い巻やより新しい巻としては扱わない

=item *
刊行の途中で巻号体系が変更になった場合（ I<45-54;1-4+> など)のケースには対応していない

=back

=head2 EXPORT

デフォルトではEXPORTは行わない。ただし、次の関数は I<use MARC::diffSerials ':all';> にてEXPORTが可能。

=over

=item *  fill_all

=item *  fill_old

=item *  fill_new

=item *  fill_middle

=item *  middle_lack

=item *  marge_vols

=back

=cut

# This allows declaration	use MARC::diffSerials ':all';
our %EXPORT_TAGS = ( 'all' => [ qw(
    middle_lack
    fill_all
    fill_old
    fill_middle
    fill_new
    marge_vols
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our $VERSION = '0.04';

=head2 FUNCTIONS

=over 4

=item fill_all(I<$vols_of_LibA, $vols_of_LibB>)

補充可能な巻を得る。第一引数は補充を受ける図書館の雑誌所蔵編集形。第二引数は雑誌を受け渡す館の雑誌所蔵編集形である。

 print MARC::diffSerials::fill_all("3(),4-6,9,10()", "2-11");
  -> 2,3(),7-9,10(),11 が出力される

=cut
sub fill_all {
    my $A_lib = shift;
    my $B_lib = shift;
    $A_lib = _check_vol_format($A_lib);
    $B_lib = _check_vol_format($B_lib);
    my @fill_all_array = ();
    my @all_vols_array = ();
    my $fill_all = "";

    my $old    = fill_old   ($A_lib, $B_lib);
    my $middle = fill_middle($A_lib, $B_lib);
    my $new    = fill_new   ($A_lib, $B_lib);

    push @all_vols_array, $old    if $old;
    push @all_vols_array, $middle if $middle;
    push @all_vols_array, $new    if $new;

    unless (@all_vols_array){
        return "";
    }
    
    my $all_vols = join ",", @all_vols_array;

    foreach my $part (_div_serials($all_vols)) {
        if ($part =~ /^\d+\-\d+$/) {
            my $first = _get_first ($part);
            my $last  = _get_last  ($part);
            foreach ($first .. $last) {
                push @fill_all_array, $_;
            }
        }
        else {
            push @fill_all_array, $part;
        }
    }
    $fill_all = _simple_checkin(@fill_all_array);
    return $fill_all;

}

=item fill_old(I<$vols_of_LibA, $vols_of_LibB>)

中途欠号のうち補充可能な古い巻を得る。第一引数は補充を受ける図書館の雑誌所蔵編集形。第二引数は雑誌を受け渡す館の雑誌所蔵編集形である。

  print MARC::diffSerials::fill_old("3(),4-6,9,10()", "2(4-5),3-9");
    -> 2(4-5) と出力される

=cut
sub fill_old {
    my $A_lib = shift;
    my $B_lib = shift;
    $A_lib = _check_vol_format($A_lib);
    $B_lib = _check_vol_format($B_lib);
    my @fill_old_array = ();
    my $fill_old = "";

    my $first_A = _get_first($A_lib);
    my $first_B = _get_first($B_lib);
    $first_A =~ s/\(.*\)$//;
    $first_B =~ s/\(.*\)$//;

    return "" if $first_A <= $first_B;

    foreach my $part (_div_serials($B_lib)) {
        my $first_part = _get_first($part); $first_part =~ s/\(.*\)$//;
        my $last_part  = _get_last ($part); $last_part  =~ s/\(.*\)$//;
        last if ($first_part >= $first_A);
        if ($last_part >= $first_A) {
             if ($part =~ /^\d+\-\d+$/) {
                  my $first_A_minos = $first_A -1;
                  my $part_first = _get_first($part);
                  if ($part_first != $first_A_minos) {
                      $part =~ s/\d+$/$first_A_minos/;
                  }
                  else {
                      $part =~ s/\-\d+$//;
                  }
                  push @fill_old_array, $part;
             }
             last;
        }
        push @fill_old_array, $part;
    }
    $fill_old = join ",", @fill_old_array;
    return $fill_old;
}

=item fill_new(I<$vols_of_LibA, $vols_of_LibB>)

途中欠号のうち補充可能な新しい巻を得る。第一引数は補充を受ける図書館の雑誌所蔵編集形。第二引数は雑誌を受け渡す館の雑誌所蔵編集形である。

  print MARC::diffSerials::fill_new("3(),4-6,9,10()", "2-9,11(3-4)");
    -> 11(3-4) と出力される

=cut
sub fill_new {
    my $A_lib = shift;
    my $B_lib = shift;
    $A_lib = _check_vol_format($A_lib);
    $B_lib = _check_vol_format($B_lib);
    my @fill_new_array = ();
    my $fill_new = "";

    my $last_A = _get_last($A_lib);
    my $last_B = _get_last($B_lib);
    $last_A =~ s/\(.*\)$//;
    $last_B =~ s/\(.*\)$//; 

    return "" if $last_A >= $last_B;

    foreach my $part (_div_serials($B_lib)) {
        my $first_part = _get_first($part); $first_part =~ s/\(.*\)//;
        my $last_part  = _get_last ($part); $last_part  =~ s/\(.*\)//; 
        if ($last_part > $last_A) {
            if ($first_part <= $last_A) {
                if ($part =~ /^\d+\-\d+$/) {
                    my $last_A_plus = $last_A +1;
                    my $part_last = _get_last($part);
                    if ($part_last != $last_A_plus) {
                        $part =~ s/^\d+/$last_A_plus/;
                    }
                    else {
                        $part =~ s/^\d+\-//;
                    }
                    push @fill_new_array, $part;
                 }
            }
            else {
                push @fill_new_array, $part;
            }
        }
    }
    $fill_new = join ",", @fill_new_array;
    return $fill_new;
}
 
=item fill_middle(I<$vols_of_LibA, $vols_of_LibB>)

途中欠号のうち補充可能な巻を得る。第一引数は補充を受ける図書館の雑誌所蔵編集形。第二引数は雑誌を受け渡す館の雑誌所蔵編集形である。

  print MARC::diffSerials::fill_middle("3(),4-6,9,10()", "5-11");
    -> 7-8,10() と出力される

=cut
sub fill_middle {
    my $A_lib = shift;
    my $B_lib = shift;
    $A_lib = _check_vol_format($A_lib);
    $B_lib = _check_vol_format($B_lib);
    my @fill_middle_arary = ();
    my $fill_middle = "";

    my $middle_lack = middle_lack($A_lib);
    return "" unless $middle_lack;
    foreach my $A_lib_part (_div_serials($middle_lack)) {
        if ($A_lib_part =~ /^\d+\-\d+$/) {
            my $first = _get_first($A_lib_part);
            my $last  = _get_last ($A_lib_part);
            foreach my $vol ($first .. $last) {
                foreach my $B_lib_part (_div_serials($B_lib)) {
                    if ($B_lib_part =~ /^\d+(\-\d+)*$/) {
                        push @fill_middle_arary, $vol
                            if (_check_include($B_lib_part, $vol));
                    }
                    elsif ($B_lib_part =~ /^$vol\(*/) {
                        push @fill_middle_arary, $B_lib_part
                            unless (_check_include($B_lib_part, $vol));
                    }
                }
            }
        } 
        else {
            my ($A_lib_part_vol) = $A_lib_part =~ /^(\d+)/;
            if (_check_include($B_lib, $A_lib_part)) {
                push @fill_middle_arary, $A_lib_part_vol;
            }
            foreach my $B_lib_part (_div_serials($B_lib)) {
                if ($B_lib_part =~ /^${A_lib_part_vol}\(.*\)$/) {
                    push @fill_middle_arary, $A_lib_part_vol . "()";
                    last;
                }
           }
        }
    }
    if (@fill_middle_arary) {
        $fill_middle = _simple_checkin(@fill_middle_arary);
        return $fill_middle;
    }
    else {
        return "";
    }
}

=item middle_lack(I<$vols>)

途中欠号を得る

  print MARC::diffSerials::middle_lack("3(),4-6,9,10()");
    -> 3(),7-8,10() と出力される

=cut
sub middle_lack {
    my $vol_string = shift;
    $vol_string = _check_vol_format($vol_string);
    my @lack_vol = (); 
    my $middle_lack = "";

    my $first_vol = _get_first($vol_string);
    my $last_vol  = _get_last($vol_string);
    $first_vol    =~ s/\(.*\)$//;
    $last_vol     =~ s/\(.*\)$//;

    LOOP:
    foreach my $vol ($first_vol .. $last_vol) {
        foreach my $part (_div_serials($vol_string)) {
           if ($part =~ /^\d+\(.*\)$/) {
               $part =~ s/\(.*\)$//;
               if ($part eq $vol) {
                   $vol .= "()";
                   push @lack_vol, $vol;
                   next LOOP;
               }
            }
            if (_check_include($part, $vol)){
                next LOOP;
            }
        }
        push @lack_vol, $vol;
    }
    $middle_lack = _simple_checkin(@lack_vol) if @lack_vol;
    return $middle_lack;
}

=item marge_vols($vols_of_LibA, $vols_of_LibB)

２館の雑誌所蔵をマージする

  print MARC::diffSerials::marge_vols("3(),4-6,9,10()", "8-9,10(4)");
    -> 3(),4-6,8-9,10() と出力される

=cut
sub marge_vols {
    my $A_lib = shift;
    my $B_lib = shift;
    $A_lib = _check_vol_format($A_lib);
    $B_lib = _check_vol_format($B_lib);
    my $marge_vols = "";

    my $fill_middle = fill_middle($A_lib, $B_lib);
    my $fill_old    = fill_old   ($A_lib, $B_lib);
    my $fill_new    = fill_new   ($A_lib, $B_lib);

    my @A_lib_array       = _div_serials($A_lib)       if $A_lib;
    my @fill_old_array    = _div_serials($fill_old)    if $fill_old;
    my @fill_middle_array = _div_serials($fill_middle) if $fill_middle;
    my @fill_new_array    = _div_serials($fill_new)    if $fill_new;
   
    my %vols = ();
    foreach my $part (@fill_old_array, 
                      @A_lib_array, 
                      @fill_middle_array, 
                      @fill_new_array) {
       if ($part =~ /^\d+\-\d+$/) {
          my $first = _get_first($part); 
          my $last  = _get_last ($part); 
          foreach ($first .. $last) {
              $vols{$_} = $_;
          }
       }
       else {
           my ($vol)   = $part =~ /^(\d+)/;
           if (defined $vols{$vol} 
               && ($vols{$vol} =~ /^\d+\(.*\)$/) && ($part =~ /^\d+\(.*\)$/) ) {
               $vols{$vol} = $vol . "()";
           }
           else {
               $vols{$vol} = $part;
           }
       }
    }
    my @middle_vols_array = ();
    foreach (sort {$a <=> $b} keys %vols) { push @middle_vols_array, $vols{$_}; }
    $marge_vols = _simple_checkin(@middle_vols_array);
    return $marge_vols;
}

#  雑誌包括所蔵形を巻のまとまりに分割する
sub _div_serials  {
    my $vols = shift;  
    my $switch = 0;
    my $keep = "";
    my @_div_serials = ();

    LOOP:
    foreach my $part (split /,/, $vols) {
        # 4(4,3,5) のような形式に対応させる
        # 4(4
        if ($part =~ /\(/ & $part !~ /\)$/) {
            $keep .= $part . ',';
            $switch = 1;
            next LOOP;
        }
        if ($switch == 1) {
            # 5)
            if ($part !~ /\(/ & $part =~ /\)$/) {
                $part = $keep . $part;
                $switch = 0;
                $keep = "";
            }
            # 3
            else {
                $keep .= $part . ',';
                next LOOP;
            }
        }
        Carp::croak "not a volume format!"
           unless ($part =~ /^\d+\(.*\)$/ or $part =~ /^\d+$/ or
                   $part =~ /^\d+\-\d+$/);
        push @_div_serials, $part;
    }
    Carp::croak "not a volume format!" if $keep;
    Carp::croak "not a volume format!" if scalar @_div_serials == 0;
    return @_div_serials;
}


# 雑誌包括所蔵形から始まりの巻号を得る
sub _get_first {
    my $vols = shift;
    my $_get_first = "";
    my @parts = _div_serials($vols);
    my $first_part = shift @parts;
    if ($first_part =~ /^\d+(\(.*\))*.*\d*(\(.*\))*$/) {
        ($_get_first) = $first_part =~ /^(\d+(\(.*\))*).*?(\d*(\(.*\))*)$/;
        return $_get_first;
    }
    else {
        Carp::croak "not a volume format!";
    }
}

# 雑誌包括所蔵形から終わりの巻号を得る
sub _get_last {
    my $vols = shift;
    my $last_vol = "";
    my @parts = _div_serials($vols);
    my $last_part = pop @parts;
    if ($last_part =~ /^\d+\-\d+$/) {
        ($last_vol) = $last_part =~ /(\d+)$/;
        return $last_vol;
    }
    elsif ($last_part =~ /\d+(\(.*\))*/) {
        return $last_part;
    }
    else {
        Carp::croak "not a volume format!";
    }
}

# 巻レベルの巻の連なりを雑誌包括形にまとめる
sub _simple_checkin {
    my @vol_array = @_;
    my $vol_string = "";
    my $status = "";

    # 現在の状態を$statusで保持する
    #
    #   S1   -> 完全巻
    #   S2   -> 不完全巻 
    #   Cont -> 継続中 
    #

    if ($vol_array[0] =~ /^\d+$/){ $status = "S1" }
    else                         { $status = "S2" }
    $vol_string = $vol_array[0];
    for (my $i = 1; $i <= $#vol_array; $i++) {
        if ($status eq "Cont") {
            if ($vol_array[$i] =~ /^\d+$/){
                if ($vol_array[$i-1] != $vol_array[$i] - 1) {
                    $vol_string .= "-" . $vol_array[$i-1];
                    $vol_string .= "," . $vol_array[$i];
                    $status = "S1";
                }
                elsif ($i == $#vol_array){
                    $vol_string .= "-" . $vol_array[$#vol_array];
                }
            }
            else {
                $vol_string .= "-" . $vol_array[$i-1];
                $vol_string .= "," . $vol_array[$i];
                $status = "S2";
            }
        }
        elsif ($status eq "S1") {
            if ($vol_array[$i] =~ /^\d+$/){
                if ($vol_array[$i-1] == $vol_array[$i] - 1) {
                    $status = "Cont";
                    $vol_string .= "-" . $vol_array[$#vol_array]
                        if $i == $#vol_array;
                }
                else {
                    $vol_string .= "," . $vol_array[$i];
                }
            }
            else {
                $vol_string .= "," . $vol_array[$i];
                $status = "S2"; 
            }
        }
        elsif ($status eq "S2") {
            $vol_string .= "," . $vol_array[$i];
            if ($vol_array[$i] =~ /^\d+$/){ $status = "S1" }
            else                          { $status = "S2" }
        }
    }
    return $vol_string;
}

=back

=cut

# ある巻が雑誌包括所蔵中に含まれるかチェックする
sub _check_include {
    my $vol_string = shift;
    my $vol        = shift;
    $vol =~ s/\(.*\)$//;
    foreach my $part (_div_serials($vol_string)){
         next if $part =~ /^\d+\(.*\)$/;
         my $first = _get_first($part);
         my $last  = _get_last ($part);
         return 1 if ($vol >= $first && $vol <= $last);
    }
    return 0;
}

# 入力データが雑誌所蔵編集形の書式に沿っているかどうかチェック
sub _check_vol_format {
    my $vol = shift;
    Carp::croak "not a volume format!" unless $vol;
    $vol =~ s/\+$//;
    $vol =~ s/\s+//g;
    Carp::croak "not support change volume!" if $vol =~ /\;/;
    Carp::croak "not a volume format!" if $vol !~ /^[\d\(\,\)\-]+$/;
    my $first = _get_first($vol); my ($first_vol) = $first =~ /(\d+)/;
    my $last  = _get_last($vol);  my ($last_vol)  = $last  =~ /(\d+)/;
    Carp::croak "first vol is larger than last vol!" if $first_vol > $last_vol;
    return $vol;
}
1;
__END__

=head1 SEE ALSO

https://mbc.dl.itc.u-tokyo.ac.jp/

=head1 AUTHOR

Akira MaedaE<lt>maeda@lib.u-tokyo.ac.jpE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008 by Akira Maeda

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
